#!/usr/bin/env python

# Be aware vdist_launcher is designed to be run as an "entry_point", i.e you
# are not supossed to execute python vdist_launcher.py but just vdist command
# after installing vdist package. If you try to run vdist_launcher.py directly
# you are going to get ImportError unless you add vdist main folder (the
# one with setup.py) to PYTHONPATH.
import concurrent.futures as futures
import contextlib
import multiprocessing
import sys
import time
import traceback
from typing import Dict, List

import vdist.console_parser as console_parser
import vdist.configuration as configuration
import vdist.defaults as defaults
import vdist.builder as builder


def _get_build_configurations(arguments: Dict[str, str]) -> Dict[str, configuration.Configuration]:
    try:
        if arguments["configuration_file"] is None:
            configurations = _load_default_configuration(arguments)
        else:
            configurations = configuration.read(arguments["configuration_file"])
    except KeyError:
        configurations = _load_default_configuration(arguments)
    return configurations


def _load_default_configuration(arguments: Dict[str, str]) -> Dict[str, configuration.Configuration]:
    arguments.name = defaults.BUILD_NAME
    _configuration = configuration.Configuration(arguments)
    configurations = {defaults.BUILD_NAME: _configuration, }
    return configurations


def run_builds(configurations: Dict[str, configuration.Configuration]) -> None:
    with futures.ProcessPoolExecutor(max_workers=multiprocessing.cpu_count()) as executor:
        workers = []
        for _configuration in configurations:
            print(f"Starting building process for {_configuration}")
            worker = executor.submit(builder.build_package, configurations[_configuration])
            workers.append(worker)
            print(f"Started building process for {_configuration}")
        print_results(workers)


def print_results(workers: List[futures.Executor]) -> None:
    for future in futures.as_completed(workers):
        files_created = future.result()
        for worker_name, files in files_created.items():
            print(f"Files created by {worker_name}:")
            for file in files:
                print(file)


@contextlib.contextmanager
def time_execution():
    start_time = time.time()
    yield
    execution_time = time.time() - start_time
    print(f"Total execution time: {execution_time} seconds")


def main(args: List=sys.argv[1:]) -> None:
    try:
        with time_execution():
            console_arguments = console_parser.parse_arguments(args)
            configurations = _get_build_configurations(console_arguments)
            run_builds(configurations)
    except Exception:
        traceback.print_exc(file=sys.stdout)
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()



import os

PYTHON_BASEDIR = '/opt'
PYTHON_VERSION = '2.7.9'
LOCAL_PROFILES_DIR = 'buildprofiles'
LOCAL_PROFILES_FILE = 'profiles.json'
VDIST_USERDIR = os.path.join(os.path.expanduser('~'), '.vdist')
BUILD_BASEDIR = "/tmp/vdist"
SCRATCH_BUILDSCRIPT_NAME = 'buildscript.sh'
SCRATCH_DIR = 'scratch'
SHARED_DIR = '/work'
PACKAGE_INSTALL_ROOT = PYTHON_BASEDIR
PACKAGE_TMP_ROOT = '/tmp'
OUTPUT_FOLDER = "./"
OUTPUT_SCRIPT = False
BUILD_NAME = "Default project"



import argparse
import os.path
from typing import Dict


def _check_is_file(_string: str) -> str:
    if os.path.isfile(_string):
        return _string
    else:
        raise argparse.ArgumentTypeError("{0} file does "
                                         "not exists.".format(_string))


# TODO: Some defaults are redundant with defaults at object creation. Fix it.
def parse_arguments(args: list=None) -> Dict[str, str]:
    arg_parser = argparse.ArgumentParser(description="A tool that lets you "
                                                     "create OS packages from "
                                                     "your Python applications "
                                                     "in a clean and self "
                                                     "contained manner.\n",
                                         epilog="Follow vdist development at: "
                                                "<https://github.com/dante-signal31/vdist>")
    # There is a bug in Python 3.5 argparser that makes that missing arguments
    # don't raise a "too few arguments". While that bug is finally fixed, there
    # is a workaround in:
    #   http://stackoverflow.com/questions/23349349/argparse-with-required-subparser
    subparsers = arg_parser.add_subparsers(help="Available modes",
                                           dest="mode")
    subparsers.required = True
    automatic_subparser = subparsers.add_parser("batch",
                                                help="Automatic configuration. "
                                                     "Parameters are going to "
                                                     " be read from a"
                                                     "configuration file.")
    automatic_subparser.add_argument("configuration_file",
                                     nargs="?",
                                     default=None,
                                     type=_check_is_file,
                                     metavar="CONFIGURATION FILENAME")
    # [TODO] Console does not get output_script right when entered from console
    # in batch mode.
    automatic_subparser.add_argument("--output_script",
                                     required=False,
                                     help="Copy build script in output folder.",
                                     action="store_const",
                                     const=True,
                                     default=False)
    manual_subparser = subparsers.add_parser("manual",
                                             help="Manual configuration. "
                                                  "Parameters are going to be "
                                                  "provided through flags.")
    manual_subparser.add_argument("-a", "--app",
                                  required=True,
                                  help="Name of application to package.",
                                  metavar="APP_NAME")
    manual_subparser.add_argument("-v", "--version",
                                  required=True,
                                  help="Version of application to package.",
                                  metavar="APP_VERSION")
    source = manual_subparser.add_mutually_exclusive_group(required=True)
    source.add_argument("-g", "--source_git",
                        help="Location of remote git repository to build from.",
                        metavar="APP_GIT")
    source.add_argument("-G", "--source_git_directory",
                        help="Location of local git repository to build from.",
                        metavar="APP_GIT_DIRECTORY")
    source.add_argument("-d", "--directory",
                        help="Location of local directory to build from.",
                        metavar="APP_DIRECTORY")
    manual_subparser.add_argument("-p", "--profile",
                                  required=True,
                                  help="Build profile",
                                  metavar="BUILD_PROFILE")
    manual_subparser.add_argument("-n", "--name",
                                  required=False,
                                  help="Build name",
                                  metavar="BUILD_NAME")
    manual_subparser.add_argument("-b", "--build_deps",
                                  required=False,
                                  nargs="*",
                                  help="Build dependencies.",
                                  metavar="BUILD_DEPENDENCIES")
    manual_subparser.add_argument("-r", "--runtime_deps",
                                  required=False,
                                  nargs="*",
                                  help="Runtime dependencies.",
                                  metavar="RUNTIME_DEPENDENCIES")
    manual_subparser.add_argument("-c", "--custom_filename",
                                  required=False,
                                  help="Custom filename for generated package.",
                                  metavar="CUSTOM_FILENAME")
    manual_subparser.add_argument("-f", "--fpm_args",
                                  required=False,
                                  help="Extra arguments for FPM. (Put text "
                                       "between quotes)",
                                  metavar="FPM_ARGS")
    manual_subparser.add_argument("-i", "--pip_args",
                                  required=False,
                                  help="Extra arguments for PIP. (Put text "
                                       "between quotes)",
                                  metavar="PIP_ARGS")
    manual_subparser.add_argument("-B", "--python_basedir",
                                  required=False,
                                  help="Base directory were python "
                                       "distribution "
                                       "that is going to be packaged is placed "
                                       "inside container. (Defaults to '/opt')",
                                  metavar="PYTHON_BASEDIR")
    manual_subparser.add_argument("-R", "--package_install_root",
                                  required=False,
                                  help="Base directory were this package "
                                       "is going to be installed in target "
                                       "system. (Defaults to 'python_basedir')",
                                  metavar="INSTALL_ROOT")
    manual_subparser.add_argument("-P", "--package_tmp_root",
                                  required=False,
                                  help="Temporal folder used in docker "
                                       "container to build your package. "
                                       "(Defaults to '/tmp')",
                                  metavar="TMP_ROOT")
    manual_subparser.add_argument("-D", "--working_dir",
                                  required=False,
                                  help="Subdirectory under your source tree "
                                       "that is to be regarded as the base "
                                       "directory",
                                  metavar="WORKING_DIR")
    manual_subparser.add_argument("-C", "--compile_python",
                                  required=False,
                                  help="Indicates Python should be fetched "
                                       "from python.org, compiled and shipped "
                                       "for you.",
                                  action="store_const",
                                  const="True",
                                  default="False")
    manual_subparser.add_argument("-V", "--python_version",
                                  required=False,
                                  help="Python version to package.",
                                  metavar="PYTHON_VERSION")
    manual_subparser.add_argument("-t", "--requirements_path",
                                  required=False,
                                  help="Path to your pip requirements file, "
                                       "relative to your project root. "
                                       "(Defaults to */requirements.txt*).",
                                  metavar="REQUIREMENTS_PATH")
    manual_subparser.add_argument("-o", "--output_folder",
                                  required=False,
                                  help="Folder where generated packages should "
                                       "be placed.",
                                  metavar="OUTPUT_FOLDER")
    manual_subparser.add_argument("--output_script",
                                  required=False,
                                  help="Copy build script in output folder.",
                                  action="store_const",
                                  const=True,
                                  default=False)

    # WARNING: Keep package scripts arguments names similar to fpm arguments for
    # scripts. Arguments names from here are directly used as fpm arguments
    # names.
    manual_subparser.add_argument("--after_install",
                                  required=False,
                                  help="A script to be run after package "
                                       "installation.",
                                  metavar="AFTER_INSTALL_SCRIPT")
    manual_subparser.add_argument("--before_install",
                                  required=False,
                                  help="A script to be run before package "
                                       "installation.",
                                  metavar="BEFORE_INSTALL_SCRIPT")
    manual_subparser.add_argument("--after_remove",
                                  required=False,
                                  help="A script to be run after package "
                                       "removal.",
                                  metavar="AFTER_REMOVE_SCRIPT")
    manual_subparser.add_argument("--before_remove",
                                  required=False,
                                  help="A script to be run before package "
                                       "removal.",
                                  metavar="BEFORE_REMOVE_SCRIPT")
    manual_subparser.add_argument("--after_upgrade",
                                  required=False,
                                  help="A script to be run after package "
                                       "upgrade.",
                                  metavar="AFTER_UPGRADE_SCRIPT")
    manual_subparser.add_argument("--before_upgrade",
                                  required=False,
                                  help="A script to be run before package "
                                       "upgrade",
                                  metavar="BEFORE_UPGRADE_SCRIPT")
    parsed_arguments = vars(arg_parser.parse_args(args))
    filtered_parser_arguments = {key: value for key, value in parsed_arguments.items()
                                 if value is not None}
    return filtered_parser_arguments



import itertools
import logging
import os
from typing import Dict, Any
import docker

import vdist.defaults as defaults


class BuildMachine(object):

    def __init__(self, image: str=None):
        self.logger = logging.getLogger('BuildMachine')
        self.image = image
        self.container = None
        self.docker_client = docker.from_env(version="auto")

    @staticmethod
    def _binds_to_shell_volumes(binds: Dict[str, str]) -> Dict[str, Dict[str, str]]:
        volumes = {k: {'bind': v, 'mode': 'rw'} for k, v in binds.items()}
        return volumes

    def launch(self, build_dir: str, extra_binds: Dict[str, str]=None) -> None:
        binds = {build_dir: defaults.SHARED_DIR}
        if extra_binds:
            binds = dict(itertools.chain(binds.items(), extra_binds.items()))
        path_to_command = os.path.join(
            defaults.SHARED_DIR,
            defaults.SCRATCH_DIR,
            defaults.SCRATCH_BUILDSCRIPT_NAME
        )
        self.container = self._start_container(binds)
        self._run_command_on_container(path_to_command)

    def _run_command_on_container(self, path_to_command: str) -> int:
        result = self.container.exec_run(path_to_command, stream=True)
        for line in result.output:
            self.logger.info(line.decode("utf8"))
        return result.exit_code

    # I've been unable to locate Container class in docker exported hierarchy. So I've set Any as return type.
    def _start_container(self, binds: Dict[str, str]) -> Any:
        self.logger.info(f'Starting container: {self.image}')
        container = self.docker_client.containers.run(image=self.image, detach=True, command="bash", tty=True,
                                                      stdin_open=True, volumes=self._binds_to_shell_volumes(binds))
        return container

    def shutdown(self) -> None:
        self._stop_container()
        self._remove_container()

    def _remove_container(self) -> None:
        self.logger.info(f'Removing container: {self.container.id}')
        self.container.remove(force=True)

    def _stop_container(self) -> None:
        self.logger.info(f'Stopping container: {self.container.id}')
        self.container.stop()



from typing import Dict


def git(uri: str=None, branch: str='master') -> Dict[str, str]:
    if uri.endswith('.git'):
        uri = uri[:-4]
    return dict(type='git', uri=uri, branch=branch)


def directory(path:str =None) -> Dict[str, str]:
    path = path.rstrip('/')
    return dict(type='directory', path=path)


def git_directory(path: str=None, branch: str='master') -> Dict[str, str]:
    path = path.rstrip('/')
    return dict(type='git_directory', path=path, branch=branch)



import logging
import os
import shutil
import re
import json
import tempfile
from typing import Tuple, Dict

from jinja2 import Environment, FileSystemLoader

import vdist.configuration as configuration
import vdist.defaults as defaults
import vdist.buildmachine as buildmachine


def build_package(_configuration: configuration.Configuration) -> dict:
    builder = _prepare_build(_configuration)
    if _configuration.output_script:
        builder.copy_script_to_output_folder(_configuration)
    builder.start_build()
    files_created = builder.move_package_to_output_folder(_configuration)
    return {_configuration.name: files_created, }


def _prepare_build(_configuration: configuration.Configuration) -> 'Builder':
    builder = _generate_builder(_configuration)
    builder.get_available_profiles()
    builder.create_build_folder_tree()
    builder.populate_build_folder_tree()
    _create_output_folder(_configuration)
    return builder


def _generate_builder(_configuration: configuration.Configuration) -> 'Builder':
    builder = Builder(process_name=_configuration.name)
    builder.add_build(**_configuration.builder_parameters)
    return builder


# TODO: Possibly redundant with already existing code. REFACTOR
def _get_script_output_filename(_configuration: configuration.Configuration) -> str:
    script_filename = f"{_get_package_folder_name(_configuration)}.sh"
    output_filepath = os.path.join(_configuration.output_folder,
                                   script_filename)
    return output_filepath


def _move_generated_package(_configuration: configuration.Configuration, package_folder: str) -> list:
    files_moved = []
    for file in os.listdir(package_folder):
        # Only copy files with extension as they are likely the generated
        # package.
        if os.path.splitext(file)[1] != "":
            source_file_pathname = os.path.join(package_folder, file)
            shutil.copy(source_file_pathname, _configuration.output_folder)
            destination_file_pathname = os.path.join(_configuration.output_folder, file)
            files_moved.append(destination_file_pathname)
    return files_moved


def _get_generated_package_folder(_configuration: configuration.Configuration, source_folder: str) -> str:
    return os.path.join(source_folder, _get_package_folder_name(_configuration))


def _create_output_folder(_configuration: configuration.Configuration) -> None:
    if not os.path.exists(_configuration.output_folder):
        _create_folder(_configuration)


def _create_folder(_configuration: configuration.Configuration) -> None:
    os.makedirs(_configuration.output_folder, exist_ok=True)


# TODO: Possibly redundant with already existing code. REFACTOR
def _get_package_folder_name(_configuration: configuration.Configuration) -> str:
    package_folder = "{app}-{version}-{profile}".format(**_configuration.builder_parameters)
    return package_folder


# Standard shutil.copytree has problems copying into an already populated
# folder, like we do with vdist. Workaround found here:
#     https://stackoverflow.com/questions/1868714/how-do-i-copy-an-entire-directory-of-files-into-an-existing-directory-using-python
def _copytree(src: str, dst: str, symlinks: bool=False, ignore: bool=None) -> None:
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            os.makedirs(os.path.dirname(d), exist_ok=True)
            shutil.copy2(s, d)


class BuildProfile(object):

    def __init__(self, **kwargs):
        self.required_attrs = ['profile_id', 'docker_image', 'script']
        # TODO: I'm not sure about insecure_registry is actually used any longer. Try to remove it.
        self.optional_attrs = ['insecure_registry']

        for arg in kwargs:
            if arg not in self.required_attrs and \
                    arg not in self.optional_attrs:
                raise AttributeError('attribute not allowed: %s' % arg)

        self.__dict__.update(kwargs)

        self.validate()

        if hasattr(self, 'insecure_registry') and \
                self.insecure_registry == 'true':
            self.insecure_registry = True
        else:
            self.insecure_registry = False

    def validate(self):
        for attr in self.required_attrs:
            if not hasattr(self, attr):
                raise AttributeError(
                    f'build profile misses attribute: {attr}')
        return True

    def __str__(self):
        return str(self.__dict__)


class Build(object):

    def __init__(self, app, version, source, profile,
                 name=None, use_local_pip_conf=False, build_deps=None,
                 runtime_deps=None, custom_filename=None,
                 fpm_args='', pip_args='',
                 package_install_root=None,
                 package_tmp_root=None,
                 working_dir='', python_basedir=None,
                 compile_python=True,
                 python_version=defaults.PYTHON_VERSION,
                 requirements_path='/requirements.txt',
                 after_install=None,
                 before_install=None,
                 after_remove=None,
                 before_remove=None,
                 after_upgrade=None,
                 before_upgrade=None):
        self.app = app
        self.version = version.format(**os.environ)
        self.source = source
        self.use_local_pip_conf = use_local_pip_conf
        if package_install_root is None:
            self.package_install_root = defaults.PACKAGE_INSTALL_ROOT.format(**os.environ)
        else:
            self.package_install_root = package_install_root.format(**os.environ)
        if package_tmp_root is None:
            self.package_tmp_root = defaults.PACKAGE_TMP_ROOT.format(**os.environ)
        else:
            self.package_tmp_root = package_tmp_root.format(**os.environ)
        self.working_dir = working_dir.format(**os.environ)
        self.requirements_path = requirements_path.format(**os.environ)
        if python_basedir is None:
            self.python_basedir = "/".join([defaults.PYTHON_BASEDIR, app]).format(**os.environ)
        else:
            self.python_basedir = python_basedir.format(**os.environ)
        self.compile_python = compile_python
        self.python_version = python_version.format(**os.environ)
        if custom_filename:
            self.custom_filename = custom_filename.format(**os.environ)
        else:
            self.custom_filename = None

        self.build_deps = []
        if build_deps:
            self.build_deps = build_deps

        self.runtime_deps = []
        if runtime_deps:
            self.runtime_deps = runtime_deps

        self.profile = profile
        # I don't like method chaining but I didn't get it to work with a
        # auxiliary variable.
        self.fpm_args = self._append_scripts_to_args(locals())
        self.pip_args = pip_args.format(**os.environ)

        if not name:
            self.name = self.get_safe_dirname()
        else:
            self.name = name

        # To be set by Builder when build process is started.
        self.build_tmp_dir = None
        self.scratch_dir = None

    def __str__(self):
        return str(self.__dict__)

    def get_project_root_from_source(self) -> str:
        if self.source['type'] == 'git':
            return os.path.basename(self.source['uri'].rstrip('/'))
        if self.source['type'] in ['directory', 'git_directory']:
            return os.path.basename(self.source['path'].rstrip('/'))
        return ''

    def get_safe_dirname(self) -> str:
        return re.sub(
            '[^A-Za-z0-9\.\-]',
            '_',
            '-'.join(
                [self.app, self.version, self.profile]
            )
        )

    @staticmethod
    def _append_scripts_to_args(arguments) -> str:
        fpm_args = arguments["fpm_args"]

        if arguments["package_tmp_root"] is None:
            tmp_root = defaults.PACKAGE_TMP_ROOT
        else:
            tmp_root = arguments["package_tmp_root"]

        for script_type in configuration.SCRIPTS_ARGUMENTS:
            if arguments[script_type] is not None:
                # fpm uses dashes in its arguments but I don't because they
                # cannot be used as dict keys. So I use underscores that
                # must be converted here.
                fpm_argument_name = script_type.replace("_", "-")
                script_path = os.path.join(tmp_root,
                                           arguments["app"],
                                           arguments[script_type])
                fpm_args = "".join(["--{key} {value} ".format(key=fpm_argument_name,
                                                              value=script_path),
                                    fpm_args])
        return fpm_args


class Builder(object):

    def __init__(
            self,
            process_name=defaults.BUILD_NAME,
            profiles_dir=defaults.LOCAL_PROFILES_DIR,
            machine_logs=True):
        logging.basicConfig(format=f'%(asctime)s %(levelname)s [{process_name}] %(name)s %(message)s',
                            level=logging.INFO)
        self.logger = logging.getLogger('Builder')

        self.build_basedir = tempfile.mkdtemp(prefix="vdist")
        self.profiles = {}
        # Actually, list of pending builds is no longer stored here, but in
        # vdist_launcher configurations when console launcher is used.
        # self.builds = []
        self.build = None

        self.machine_logs = machine_logs
        self.local_profiles_dir = profiles_dir
        self._load_profiles()

    def add_build(self, **kwargs) -> None:
        self.build = Build(**kwargs)

    # TODO: Possibly redundant with already existing code. REFACTOR
    # def _create_vdist_dir(self) -> None:
    #     vdist_path = os.path.join(os.path.expanduser('~'), '.vdist')
    #     if not os.path.exists(vdist_path):
    #         self.logger.info(f'Creating: {vdist_path}')
    #         os.mkdir(vdist_path)

    def _add_profiles_from_file(self, config_file) -> None:
        with open(config_file) as f:
            profiles = json.loads(f.read())

            for profile_id in profiles:
                profile = BuildProfile(
                    profile_id=profile_id,
                    docker_image=profiles[profile_id]['docker_image'],
                    script=profiles[profile_id]['script'],
                    insecure_registry=profiles[profile_id].get(
                        'insecure_registry', 'false')
                )

                self.profiles[profile_id] = profile

    def _load_profiles(self) -> None:
        internal_profiles = os.path.join(
            os.path.dirname(__file__),
            'profiles', 'internal_profiles.json')
        self._add_profiles_from_file(internal_profiles)

        local_profiles = os.path.join(
            self.local_profiles_dir, defaults.LOCAL_PROFILES_FILE)
        if os.path.isfile(local_profiles):
            self._add_profiles_from_file(local_profiles)

    def _render_template(self, build) -> None:
        internal_template_dir = os.path.join(
            os.path.dirname(__file__), 'profiles')

        local_template_dir = os.path.abspath(self.local_profiles_dir)

        env = Environment(loader=FileSystemLoader([internal_template_dir,
                                                   local_template_dir]))

        if build.profile not in self.profiles:
            raise BuildProfileNotFoundException(
                f'profile not found: {build.profile}')

        profile = self.profiles[build.profile]
        template_name = profile.script
        template = env.get_template(template_name)

        # local uid and gid are needed to correctly set permissions
        # on the created artifacts after the build completes
        return template.render(
            local_uid=os.getuid(),
            local_gid=os.getgid(),
            project_root=build.get_project_root_from_source(),
            shared_dir=defaults.SHARED_DIR,
            scratch_folder_name=defaults.SCRATCH_DIR,
            **build.__dict__
        )

    def _clean_build_basedir(self) -> None:
        if os.path.exists(self.build_basedir):
            shutil.rmtree(self.build_basedir)

    def _create_build_basedir(self) -> None:
        os.mkdir(self.build_basedir)

    @staticmethod
    def _write_build_script(path, script) -> None:
        with open(path, 'w+') as f:
            f.write(script)
        os.chmod(path, 0o777)

    def _populate_scratch_dir(self, build) -> None:
        # write rendered build script to scratch dir
        self._write_build_script(
            os.path.join(build.scratch_dir, defaults.SCRATCH_BUILDSCRIPT_NAME),
            self._render_template(build)
        )

        # copy local ~/.pip if necessary
        if build.use_local_pip_conf:
            _copytree(
                os.path.join(os.path.expanduser('~'), '.pip'),
                os.path.join(build.scratch_dir, '.pip')
            )

        # local source type, copy local dir to scratch dir
        if build.source['type'] in ['directory', 'git_directory']:
            if not os.path.exists(build.source['path']):
                raise ValueError(
                    f'path does not exist: {build.source["path"]}')
            else:
                subdir = os.path.basename(build.source['path'])
                _copytree(
                    build.source['path'].rstrip('/'),
                    os.path.join(build.scratch_dir, subdir)
                )

    def _create_build_dir(self, build) -> Tuple[str, str]:
        subdir_name = build.get_safe_dirname()

        build_dir = os.path.join(self.build_basedir, subdir_name)

        if os.path.exists(build_dir):
            shutil.rmtree(build_dir)

        os.mkdir(build_dir)

        # create "scratch" subdirectory for stuff needed at build time
        scratch_dir = os.path.join(build_dir, defaults.SCRATCH_DIR)
        os.mkdir(scratch_dir)

        return build_dir, scratch_dir

    def run_build(self) -> None:
        profile = self.profiles[self.build.profile]

        self.logger.info(f'launching docker image: {profile.docker_image}')

        build_machine = buildmachine.BuildMachine(
            image=profile.docker_image
        )

        self.logger.info(f'Running build machine for: {self.build.name}')
        build_machine.launch(build_dir=self.build.build_tmp_dir)

        self.logger.info(f'Shutting down build machine: {self.build.name}')
        build_machine.shutdown()

        self.logger.info(f'*** Resulting OS packages are in: {self.build.build_tmp_dir} ***')

    def get_available_profiles(self) -> Dict[str, BuildProfile]:
        self._load_profiles()
        return self.profiles

    def start_build(self) -> None:
        # self._create_vdist_dir()
        if self.build is None:
            raise NoBuildsFoundException()
        self.run_build()

    def populate_build_folder_tree(self) -> None:
        self._populate_scratch_dir(self.build)

    def create_build_folder_tree(self) -> None:
        self._start_build_basedir()
        self._start_build_folders()

    def _start_build_basedir(self) -> None:
        self._clean_build_basedir()
        self._create_build_basedir()

    def _start_build_folders(self) -> None:
        build_tmp_dir, scratch_dir = self._create_build_dir(self.build)
        self.build.build_tmp_dir = build_tmp_dir
        self.build.scratch_dir = scratch_dir

    def copy_script_to_output_folder(self, _configuration) -> None:
        source_folder = self.build.scratch_dir
        script_filepath = os.path.join(source_folder,
                                       defaults.SCRATCH_BUILDSCRIPT_NAME)
        script_output_filepath = _get_script_output_filename(_configuration)
        shutil.copy(script_filepath, script_output_filepath)

    def move_package_to_output_folder(self, _configuration: configuration.Configuration) -> list:
        files_moved = _move_generated_package(_configuration, self.build.build_tmp_dir)
        return files_moved


class BuildProfileNotFoundException(Exception):
    pass


class TemplateNotFoundException(Exception):
    pass


class NoBuildsFoundException(Exception):
    pass



from glob import glob
import os
from setuptools import setup, find_packages
from typing import List
import ci_scripts.ci_tools as tools

VDIST_PACKAGES_CONFIG = "packaging/vdist_build.cnf"


def find_folders_with_this_name(dir_name):
    """ Look for folder with given name, searching from current working dir.

    :param dir_name: Folder name to look for.
    :return: Relative path, from current working dir, where folder is.
    """
    for dir, dirs, files in os.walk('.'):
        if dir_name in dirs:
            yield os.path.relpath(os.path.join(dir, dir_name))


def find_man_pages():
    """ Look for every folder named "man", inside current working dir, and include
    its files relatives paths in a list suitable to be passed to a data_files
    setup.py parameter. This list will set those manpages to be installed to
    "share/man/man<section>" at target box.
    This code was inspired by a really interesting answer to:
    https://github.com/pypa/packaging-problems/issues/72#issuecomment-279162312

    :return: A list with tuples ("share/man/man<section>", local_relative_path_to_manpage)
    """
    data_files = []
    man_sections = {}
    for dir in find_folders_with_this_name('man'):
        for file in os.listdir(dir):
            section = file.split('.')[-1]
            man_sections[section] = man_sections.get(section, []) + [os.path.join(dir, file)]
    for section in man_sections:
        data_files.append(('share/man/man'+section, man_sections[section]))
    return data_files


def find_info_pages():
    """ Look for every folder named "info", inside current working dir, and include
    its files relatives paths in a list suitable to be passed to a data_files
    setup.py parameter. This list will set those infopages to be installed to
    "share/info" at target box.
    This code was inspired by a really interesting answer to:
    https://github.com/pypa/packaging-problems/issues/72#issuecomment-279162312

    :return: A list with tuples ("share/info", local_relative_path_to_infopage)
    """
    data_files = []
    info_pages = {}
    for dir in find_folders_with_this_name('info'):
        for file in glob(os.path.join(dir, '*.info')):
            info_pages[dir] = info_pages.get(dir, []) + [file]
    for dir in info_pages:
        data_files.append(('share/info', info_pages[dir]))
    return data_files


setup(
    name='vdist',
    # Version is set in packaging/vdist_build.cnf and returned through
    # tools.get_current_version().
    version=tools.get_current_version(VDIST_PACKAGES_CONFIG),
    description='Create OS packages from Python '
                'projects using Docker containers',
    long_description='Create OS packages from Python '
                     'projects using Docker containers',
    author='objectified, dante-signal31',
    author_email='objectified@gmail.com, dante.signal31@gmail.com',
    license='MIT',
    url='https://github.com/dante-signal31/vdist',
    data_files=find_man_pages(),
    packages=find_packages(exclude=["tests", "integration-tests",
                                    "ci_scripts", "examples", "docs"]),
    install_requires=['jinja2==2.7.3', 'docker==3.2.1'],
    entry_points={'console_scripts': ['vdist=vdist.vdist_launcher:main', ], },
    package_data={'': ['internal_profiles.json', '*.sh']},
    tests_require=['pytest'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    zip_safe=False,
    keywords='python docker deployment packaging',
)



import configparser
from typing import Dict

import vdist.defaults as defaults
import vdist.source as source

LISTABLE_ARGUMENTS = {"source_git", "source_git_directory", "runtime_deps",
                      "build_deps"}
LONG_TEXT_ARGUMENTS = {"fpm_args", "pip_args"}
PROCESSABLE_ARGUMENTS = {"source_directory", "compile_python",
                         "fpm_args"}
SCRIPTS_ARGUMENTS = {"after_install", "before_install", "after_remove",
                     "before_remove", "after_upgrade", "before_upgrade"}
USELESS_ARGUMENTS = {"mode", "name", "output_folder", "output_script"}
PROCESSABLE_ARGUMENTS |= LISTABLE_ARGUMENTS
PROCESSABLE_ARGUMENTS |= LONG_TEXT_ARGUMENTS


class Configuration(object):

    # Using a fixed set of parameters for __init__ might be cleaner here but
    # I've found it less scalable in this specific case. After many test I
    # feel better for scalation using a generic dir as input and process it
    # using various arguments sets, it keeps my __init_ shorter and it is easier
    # add and remove arguments.
    def __init__(self, arguments):
        self.output_folder = arguments.get("output_folder",
                                           defaults.OUTPUT_FOLDER)
        self.output_script = arguments.get("output_script",
                                           defaults.OUTPUT_SCRIPT)
        self.name = arguments.get("name", defaults.BUILD_NAME)
        self.builder_parameters = {key: value for key, value in arguments.items()
                                   if key not in PROCESSABLE_ARGUMENTS and
                                   key not in USELESS_ARGUMENTS}
        self._process_listable_arguments(arguments)
        self._process_long_text_arguments(arguments)
        self._process_source_directory_argument(arguments)
        self._process_compile_python_argument(arguments)

    def _process_source_directory_argument(self, arguments):
        if "source_directory" in arguments.keys():
            directory = arguments["source_directory"]
            directory = directory.strip()
            self.builder_parameters["source"] = source.directory(directory)

    def _process_compile_python_argument(self, arguments):
        if "compile_python" in arguments.keys():
            self.builder_parameters["compile_python"] = bool(
                arguments["compile_python"])

    def _process_listable_arguments(self, arguments):
        argument_keys = set(arguments.keys())
        listable_arguments_found = LISTABLE_ARGUMENTS.intersection(argument_keys)
        for argument in listable_arguments_found:
            generated_list = _create_list(arguments[argument])
            if argument == "source_git":
                self.builder_parameters["source"] = source.git(
                    generated_list[0],
                    generated_list[1])
            if argument == "source_git_directory":
                self.builder_parameters["source"] = source.git_directory(
                    generated_list[0],
                    generated_list[1])
            if argument in ["runtime_deps", "build_deps"]:
                self.builder_parameters[argument] = generated_list

    def _process_long_text_arguments(self, arguments):
        argument_keys = set(arguments.keys())
        long_text_arguments_found = LONG_TEXT_ARGUMENTS.intersection(
            argument_keys)
        for argument in long_text_arguments_found:
            text_with_no_cr = _remove_cr(arguments[argument])
            self.builder_parameters[argument] = text_with_no_cr


def _create_list(_list):
    return _list if isinstance(_list, list) else [element.strip()
                                                  for element in _list.split(",")]


def _remove_cr(text):
    # Removes carriage returns from given text.
    # Great reference:
    #   http://stackoverflow.com/questions/3939361/remove-specific-characters-from-a-string-in-python
    #   http://stackoverflow.com/questions/21038891/what-does-table-in-the-string-translate-function-mean
    return text.translate({ord('\n'): ord(' '), })


def read(configuration_file: str) -> Dict[str, Configuration]:
    # Should return a dict whose keys should be configfile sections
    # (except DEFAULT) and configurations objects as values.
    parser = _get_config_parser()
    parser.read(configuration_file)
    configurations = {}
    for section in parser.sections():
        parameters = _get_section_values(parser, section)
        parameters["name"] = section
        configurations[section] = Configuration(parameters)
    return configurations


def _get_section_values(parser, section):
    section_values = {key: parser[section][key]
                      for key in parser[section]}
    return section_values


def _get_config_parser():
    parser = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())
    return parser

